using System;
using System.Data;
using System.Collections;
using System.ComponentModel;

using DataBinder.CrmSdk;
using DataBinder.MetadataServiceSdk;



namespace DataBinder
{
    /// <summary>
    /// Provides an IList that can be bound directly to a standard .NET Data Grid
    /// </summary>
    public class UniversalEntityList : IList
    {
        private BusinessEntity[] _entities;
        PropertyDescriptorCollection _columnDescriptors;

        public UniversalEntityList(BusinessEntityCollection crmEntities)
        {
            // Standard MetaData Service Setup
            MetadataService service = new MetadataService();
            service.Credentials = System.Net.CredentialCache.DefaultCredentials;

            // Get the name of the entity
            // Note: We have to convert this to lower because RetrieveEntityMetadata expects the schema name
            string entityName = crmEntities.EntityName;
            entityName = entityName.ToLower();

            // Get the metadata for the currently list's entity
            // This metadata is used to create a "Property Descriptor Collection"
            EntityMetadata entityData = service.RetrieveEntityMetadata(entityName, EntityFlags.EntityOnly | EntityFlags.IncludeAttributes);

            // Cache locals
            _columnDescriptors = UniversalEntity.CreatePropertyDescriptorCollection(entityData.Attributes);
            _entities = crmEntities.BusinessEntities;
        }

        #region UniversalEntityListEnumerator Class

        /// <summary>
        /// Enumerator for the Universal Entity List
        /// </summary>
        class UniversalEntityListEnumerator : IEnumerator
        {
            private UniversalEntityList _entityList;
            private int _index = -1;

            internal UniversalEntityListEnumerator(UniversalEntityList entityList)
            {
                _entityList = entityList;
            }

            public bool MoveNext()
            {
                // If there are no records or moving forward would go past the total
                // number of available records, return false.
                if (_entityList._entities.Length == 0 || (_index + 1) >= _entityList._entities.Length)
                {
                    return false;
                }

                _index++;
                return true;
            }

            public void Reset()
            {
                _index = -1;
            }

            public virtual Object Current
            {
                get
                {
                    return _entityList[_index];
                }
            }
        }

        #endregion

        #region IList Members

        /// <summary>
        /// Read-only indexer that provides access to the list's Universal Entities
        /// </summary>
        public object this[int index]
        {
            get
            {
                DynamicEntity crmEntity = (DynamicEntity)_entities[index];

                UniversalEntity entity = UniversalEntity.CreateFromDynamicEntity(crmEntity, _columnDescriptors);

                return (object)entity;
            }
            set
            {
                throw new NotImplementedException("This method is not supported in this implementation");
            }
        }

        /// <summary>
        /// The total number of entities in the list
        /// </summary>
        public int Count
        {
            get { return _entities.Length; }
        }

        public bool IsFixedSize
        {
            get { return true; }
        }

        public bool IsReadOnly
        {
            get { return true; }
        }
        public virtual Object SyncRoot
        {
            get { return this; }
        }

        public virtual bool IsSynchronized
        {
            get { return false; }
        }

        public virtual IEnumerator GetEnumerator()
        {
            return new UniversalEntityListEnumerator(this);
        }

        #region Unsupported Features

        public void CopyTo(Array array, int index)
        {
            throw new NotImplementedException("This method is not supported in this implementation");
        }

        public int Add(object value)
        {
            throw new NotImplementedException("This method is not supported in this implementation");
        }

        public int IndexOf(object value)
        {
            throw new NotImplementedException("This method is not supported in this implementation");
        }

        public void Insert(int index, object value)
        {
            throw new NotImplementedException("This method is not supported in this implementation");
        }

        public void Remove(object value)
        {
            throw new NotImplementedException("This method is not supported in this implementation");
        }

        public void RemoveAt(int value)
        {
            throw new NotImplementedException("This method is not supported in this implementation");
        }

        public bool Contains(object value)
        {
            throw new NotImplementedException("This method is not supported in this implementation");
        }

        public void Clear()
        {
            throw new NotImplementedException("This method is not supported in this implementation");
        }

        #endregion

        #endregion
    }
}
